/*-
|> File
|| pserver.c
||
|> Description
||
|> Comments
||
|> History
|| 95045 _mf Created.
||
||==================================================================|>
*/
#ifndef _NOIDENT
#ident "$Id: pserver.c,v 2.11 1995/11/14 04:01:40 mreynold Exp $"
#endif

#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>

#define _POSIX_C_SOURCE 5 /* because signal.h is stupid */

#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>

#define NEED_REPLIES    /* because Xproto.h is stupid */
#include <X11/Xproto.h>
#include <X11/Xlib.h>

#include "tcpip.h"
#include "printer.h"

#define _DECLARE_GLOBAL
#include "global.h"

#define MAXREQSZ    262140
#define RESOURCE_ID_MASK  0x3FFFFF
#define CLIENTOFFSET    22
#define VENBYTES    16
#define STATICMAXFD   32

/* prototypes that aren't properly defined on suns. */

int             getopt (int argc, char *const * argv, const char *optstring);
char           *strdup (const char *);

Display        *THE_DISPLAY;
struct PrintClient PCS[STATICMAXFD];
char          **fontlist;
extern char    *optarg;

typedef union xmsg
{
  xReq            req;
  unsigned char   buf[MAXREQSZ];
}

                xmsgU;

typedef union xreply
{
  xGenericReply   reply;
  unsigned char   buf[MAXREQSZ];
}

                xreplyU;

int             debug = 0;

printOptionsRec PRINT_OPTIONS;

void            initClient (struct PrintClient * pcs);

/*-=================================================================|>
||
|> Local functions
||
||==================================================================|>
||
*/

/* remove the cmap structures associated with a client */
static void
removecmaps (
       const int fd,
       const cmapinfoTablePtr table)

{
  int             ii = 0,
                  jj = 0;

  for (; ii < table->numCmap; ii++)
    if (table->cmapTable[ii].fd != fd)
      table->cmapTable[jj++] = table->cmapTable[ii];

  table->numCmap = jj;
}

/* remove the GC structures associate with a client */
static void
removegcs (
     const int fd,
     const gcinfoTablePtr table)

{
  int             ii = 0,
                  jj = 0;

  for (; ii < table->numGc; ii++)
    if (table->gcTable[ii].fd != fd)
      table->gcTable[jj++] = table->gcTable[ii];

  table->numGc = jj;
}

/* remove the window structures associated with a client */

static void
removewins (
      const int fd,
      const winfoTablePtr table)

{
  int             ii = 0,
                  jj = 0;

  for (; ii < table->numWin; ii++)
    if (table->winTable[ii].fd != fd)
      table->winTable[jj++] = table->winTable[ii];

  table->numWin = jj;
}

/* clean up the structures associated with a client */

static void
removeClient (const int fd)
{
  removewins (fd, winfoTable);
  removegcs (fd, gcinfoTable);
  removecmaps (fd, cmapinfoTable);
}

/* initialize the connection with the client. Some of these are wild guesses */

int
initConnection (int clientFd)
{
  xConnClientPrefix prefixMsg;
  xConnSetupPrefix setupMsg;
  xConnSetup      connSetupMsg;
  xPixmapFormat   pixFormatMsg;
  xWindowRoot     windowMsg;
  xDepth          depthMsg;
  xVisualType     visualMsg;
  unsigned char  *authMsg;
  int             authLen;
  char            vendor[VENBYTES];

  /* read the prefix message from the client */

  if (readLoop (clientFd, (char *) &prefixMsg,
          sizeof (prefixMsg)) < 0)
  {
    perror ("prefix read failed");
    return (False);
  }
  /* read the authorization information if there is any */

  authLen = prefixMsg.nbytesAuthProto + prefixMsg.nbytesAuthString;
  if (authLen > 0)
  {
    authMsg = (unsigned char *) malloc (authLen);
    if (readLoop (clientFd, authMsg, authLen) < 0)
    {
      perror ("auth read failed");
      free (authMsg);
      return (False);
    }
    free (authMsg);
  }
  /* write the setup message to the client */

  setupMsg.success = True;
  setupMsg.lengthReason = 0;
  setupMsg.majorVersion = 11;
  setupMsg.minorVersion = 0;
  setupMsg.length = 8 + 2 * 2 +
    (VENBYTES + sizeof (xWindowRoot) +
     2 * (sizeof (xDepth) + sizeof (xVisualType))) / 4;
  if (write (clientFd, &setupMsg, sizeof (setupMsg)) < 0)
  {
    perror ("setup write failed");
    return (False);
  }
  /* write the connection setup  message to the client */

  connSetupMsg.release = 1;
  connSetupMsg.ridBase = clientFd << CLIENTOFFSET;
  connSetupMsg.ridMask = RESOURCE_ID_MASK;
  connSetupMsg.motionBufferSize = 256;
  connSetupMsg.nbytesVendor = VENBYTES;
  connSetupMsg.maxRequestSize = MAXREQSZ / 4;
  connSetupMsg.numRoots = 1;
  connSetupMsg.numFormats = 2;
  connSetupMsg.imageByteOrder = MSBFirst;
  connSetupMsg.bitmapBitOrder = MSBFirst;
  connSetupMsg.bitmapScanlineUnit = 8;
  connSetupMsg.bitmapScanlinePad = 32;
  connSetupMsg.minKeyCode = 8;
  connSetupMsg.maxKeyCode = 132;
  if (debug)
  {
    printf ("rid base = %ld, mask = 0x%lx, maxrequest = %d\n",
      connSetupMsg.ridBase, connSetupMsg.ridMask,
      connSetupMsg.maxRequestSize);
  }
  if (write (clientFd, &connSetupMsg, sizeof (connSetupMsg)) < 0)
  {
    perror ("conn setup message write failed");
    return (False);
  }
  /* write out the vendor string */

  strcpy (vendor, "print server");
  if (write (clientFd, vendor, sizeof (vendor)) < 0)
  {
    perror ("vendor write failed");
    return (False);
  }
  /* write the pixmap formats */

  pixFormatMsg.depth = 1;
  pixFormatMsg.bitsPerPixel = 1;
  pixFormatMsg.scanLinePad = 32;
  if (write (clientFd, &pixFormatMsg, sizeof (pixFormatMsg)) < 0)
  {
    perror ("pixmap format message write failed");
    return (False);
  }
  pixFormatMsg.depth = 8;
  pixFormatMsg.bitsPerPixel = 8;
  pixFormatMsg.scanLinePad = 32;
  if (write (clientFd, &pixFormatMsg, sizeof (pixFormatMsg)) < 0)
  {
    perror ("pixmap format message write failed");
    return (False);
  }
  /* write the message describing our hypothetical screen and root window */

  windowMsg.windowId = 0;
  windowMsg.defaultColormap =
    DefaultColormap (THE_DISPLAY, DefaultScreen (THE_DISPLAY));

  windowMsg.whitePixel = 0;
  windowMsg.blackPixel = 1;
  windowMsg.currentInputMask = 0;
  windowMsg.pixWidth = 1024;
  windowMsg.pixHeight = 1024;
  windowMsg.mmWidth = 256;
  windowMsg.mmHeight = 256;
  windowMsg.minInstalledMaps = 1;
  windowMsg.maxInstalledMaps = 12;
  windowMsg.rootVisualID = PSEUDOCOLOR_VISUAL;
  windowMsg.backingStore = False;
  windowMsg.saveUnders = False;
  windowMsg.rootDepth = 8;
  windowMsg.nDepths = 2;
  if (write (clientFd, &windowMsg, sizeof (windowMsg)) < 0)
  {
    perror ("window message write failed");
    return (False);
  }
  /* write the depth information for the 8-bit PseudoColor visual */

  depthMsg.depth = 8;
  depthMsg.nVisuals = 1;
  if (write (clientFd, &depthMsg, sizeof (depthMsg)) < 0)
  {
    perror ("window message write failed");
    return (False);
  }
  /* write the 8-bit PseudoColor visual */

  visualMsg.visualID = PSEUDOCOLOR_VISUAL;
  visualMsg.class = PseudoColor;
  visualMsg.bitsPerRGB = 8;
  visualMsg.redMask = 0;
  visualMsg.greenMask = 0;
  visualMsg.blueMask = 0;
  visualMsg.colormapEntries = 256;
  if (write (clientFd, &visualMsg, sizeof (visualMsg)) < 0)
  {
    perror ("window message write failed");
    return (False);
  }
  /* write the depth information for the 24-bit TrueColor Visual */

  depthMsg.depth = 24;
  depthMsg.nVisuals = 1;
  if (write (clientFd, &depthMsg, sizeof (depthMsg)) < 0)
  {
    perror ("window message write failed");
    return (False);
  }
  /* write the 24-bit TrueColor Visual */

  visualMsg.visualID = TRUECOLOR_VISUAL;
  visualMsg.class = TrueColor;
  visualMsg.bitsPerRGB = 8;
  visualMsg.redMask = 0xff;
  visualMsg.greenMask = 0xff00;
  visualMsg.blueMask = 0xff0000;
  visualMsg.colormapEntries = 0;
  if (write (clientFd, &visualMsg, sizeof (visualMsg)) < 0)
  {
    perror ("window message write failed");
    return (False);
  }
  return (True);
}

/* call read in a loop, to make sure we read everything we're supposed to */

int
readLoop (int fd,
    char *msg,
    int length)
{
  int             nbytes,
                  total = 0;

  do
  {
    nbytes = read (fd, msg, length);
    if (nbytes <= 0)
    {
      return (nbytes);
    } else
    {
      length -= nbytes;
      total += nbytes;
      msg += nbytes;
    }
  }
  while (length > 0);
  if (debug)
    printf ("read %d bytes\n", total);
  return (total);
}

/* read a NULL-terminated vector of strings from a file */

char          **
readStrvec (char *filename)
{
  FILE           *fp;
  char            buffer[512];
  int             lineno = 1;
  char          **vector,
                **vecPtr;

  /* read it to find the number of lines */

  if ((fp = fopen (filename, "r")) == NULL)
  {
    return (NULL);
  }
  do
  {
    if (fgets (buffer, sizeof (buffer), fp) == NULL)
    {
      break;
    }
    lineno++;
  }
  while (TRUE);

  /* rewind and allocate the vector */

  rewind (fp);
  vector = vecPtr = (char **) malloc (sizeof (char *) * (lineno + 1));

  /* read the lines and "strdup" them after getting rid of the \n */

  do
  {
    if (fgets (buffer, sizeof (buffer), fp) == NULL)
    {
      break;
    }
    buffer[strlen (buffer) - 1] = '\0';
    *vecPtr++ = strdup (buffer);
  }
  while (TRUE);
  *vecPtr = NULL;

  /* close and return */

  fclose (fp);
  return (vector);
}

/* close all the resources associated with a client program */

int closeClient( int	fd)
{
    int	ii, count;

/* remove all windows associated with this file descriptor from the window
   table */

    count = 0;
    for ( ii = 0; ii < winfoTable->numWin; ii++) {
	if ( winfoTable->winTable[ ii].fd != fd) {
	    winfoTable->winTable[ count++] = winfoTable->winTable[ ii];
	}
    }
    winfoTable->numWin = count;

/* remove all colormaps associated with this fd from the colormap table */

    count = 0;
    for ( ii = 0; ii < cmapinfoTable->numCmap; ii++) {
	if ( cmapinfoTable->cmapTable[ ii].fd != fd) {
	    cmapinfoTable->cmapTable[ count++] = cmapinfoTable->cmapTable[ ii];
	}
    } 
    cmapinfoTable->numCmap = count;

    /* remove all the graphics contexts associated with this fd */

    count = 0;
    for ( ii = 0; ii < gcinfoTable->numGc; ii++) {
	if ( gcinfoTable->gcTable[ ii].fd != fd) {
	    gcinfoTable->gcTable[ count++] = gcinfoTable->gcTable[ ii];
	}   
    }
    gcinfoTable->numGc = count;

    /* remove all the fonts associated with this fd */
	
    count = 0;
    for ( ii = 0; ii < fontTable->numFonts; ii++) {
	if ( fontTable->fontTable[ ii].fd != fd) {
	    fontTable->fontTable[ count++] = fontTable->fontTable[ ii];
	}   
    }
    fontTable->numFonts = count;
    return( TRUE);
}	   

/* top level event loop based on select call to the client programs fd's */

static void
eventLoop (
     fd_set readfds,
     fd_set writefds,
     fd_set exceptfds,
     fd_set readmask,
     int acceptFd,
     int highestFd)

{
  xmsgU           message;
  xreplyU         reply;
  int             clientFd,
                  clientFlag,
                  replyLen;

  for (;;)
  {
    /* copy to the select mask */
    {
      int             ii = 0;

      for (; ii <= highestFd; ii++)
        if (FD_ISSET (ii, &readfds))
          FD_SET (ii, &readmask);
    }

/* also copy the display connection to the select mask for read and write */

    FD_SET( XConnectionNumber( THE_DISPLAY), &readmask);

    if (select (highestFd + 1, &readmask, &writefds, &exceptfds, NULL) < 0)
    {
      /* some client disconnected */
      if (errno == EBADF)
      {
      /* if the X server disconnected, then we shut down */
 
	if ( FD_ISSET( XConnectionNumber( THE_DISPLAY), &readmask)) {
	  fprintf( stderr, "display server closed connection");
	  exit( -1);
	}
        if (debug)
          printf ("client closed connection\n");
        for (clientFd = 0; clientFd <= highestFd; clientFd++)
        {
          if (clientFd == acceptFd)
            continue;

          if (FD_ISSET (clientFd, &readmask)) {
	      FD_CLR (clientFd, &readfds);
	      closeClient( clientFd);
	  }	      
        }
      }
      /* if there's an input on the accept socket, accept the
         connection */
    } else if (FD_ISSET (acceptFd, &readmask))
    {
      clientFd = acceptConnection (acceptFd);
      if (debug)
        printf ("accepting connection\n");

      /* do the "init" conversation right away, since it's a
         little different than the rest */

      if (initConnection (clientFd))
      {
        FD_SET (clientFd, &readmask);

        /* once the client has connected, we initialize
           the fields in the PrintClient structure for
           that file descriptor */

        initClient (&PCS[clientFd]);
        if (clientFd > highestFd)
          highestFd = clientFd;
      }
      /* otherwise, it's just input from the client, and a
         response */

    } else
    {
      for (clientFd = 0; clientFd <= highestFd; clientFd++)
      {
        if (clientFd == acceptFd)
          continue;

        if (FD_ISSET (clientFd, &readmask))
        {
          clientFlag = True;

          /* read the request, find out how big
             it is, and read the rest of it, then
             send the reply if there is one */

          if (readLoop (clientFd, (char *) &message.req,
               sizeof (message.req)) <= 0)
          {
            if (debug)
              printf ("client %d disconnected\n", clientFd);
            clientFlag = False;
          }
          if (message.req.length > 1)
          {
            if (readLoop (clientFd,
                    (char *) message.buf + sizeof (xReq),
               (message.req.length << 2) -
                sizeof (xReq)) <= 0)
            {
              if (debug)
                printf ("client %d disconnected\n", clientFd);
              clientFlag = False;
            }
          }
          if (debug)
            printf ("reading request %d words\n", message.req.length);
          DecodeRequest (clientFd, message.buf, reply.buf,
                   &replyLen);
          if (replyLen > 0)
          {
            /* if the reply length is less
               than the length of a generic
               reply, then pump up the
               volume */
            if (replyLen < sizeof (xGenericReply))
              replyLen = sizeof (xGenericReply);

            if (debug)
              printf ("sending reply %d words\n", replyLen >> 2);

            /* if the reply write failed,
               then the client disconnected */
            if (write (clientFd, reply.buf, replyLen) < 0)
            {
              if (debug)
                printf ("client %d disconnected\n", clientFd);
              clientFlag = False;
            }
          }
          /* if anything failed, then close out
             the client, and remove the
             associated structures */

          if (!clientFlag)
          {
            close (clientFd);
            removeClient (clientFd);
            FD_CLR (clientFd, &readmask);
          }
        }
      }
    }
  }
}

/*-=================================================================|>
||
|> Public functions
||
||==================================================================|>
||
|> Function
|| main
||
|> Description
|| system entry point to the Print Server.
||
|> Comments
||
||
||==================================================================|>
*/
int
main (
      int argc,
      char *argv[],
      char *envT[])

{
  int             port = X_TCP_PORT + 1,
                  acceptFd,
                  highestFd,
                  c;
  fd_set          readfds,
                  writefds,
                  exceptfds,
                  readmask;
  char           *dispName = getenv ("DISPLAY");
  FILE           *fp;
  struct sigaction act,
                  oact;
  char           *printFonts = getenv ("XPRINTFONTS"),
                 *printOpts = getenv ("XPRINTOPTS");

  /* options: -p <port #>      port # for server -d <display-name> name
     of real display to connect to -g         debug flag on/off
  
  set the defaults: */

  if (!dispName)
    dispName = "unix:0.0";

  /* open and read the font file */
  if (!printFonts)
  {
    fprintf (stderr, "You must set the environment variable XPRINTFONTS.\n");
    return -1;
  }
  /* parse the arguments */
  while ((c = getopt (argc, argv, "gp:d:")) != EOF)
    switch (c)
    {
    case 'g':
      debug = TRUE;
      break;
    case 'p':
      port = X_TCP_PORT + atoi (optarg);
      break;
    case 'd':
      dispName = strdup (optarg);
      break;
    default:
      fprintf (stderr,
         "Usage %s [-p port#][-d display][ -g]\n", argv[0]);
      exit (-1);
      break;
    }

  /* try to open the display to the real server */
  if (!(THE_DISPLAY = XOpenDisplay (dispName)))
  {
    fprintf (stderr, "Failed to open display %s.\n", dispName);
    return -2;
  }

  if (!(fontlist = readStrvec (printFonts)))
  {
    fprintf (stderr, "Failed to open font mapping file %s.\n", printFonts);
    return -3;
  }
  /* initialize the tables for colormaps, GC's, windows, etc. */
  initTables ();

  /* set the defaults, and read the options file if there is one */
  PRINT_OPTIONS.indexCspaceSupport = TRUE;
  if (!printOpts)
  {
    if (debug)
      fprintf (stderr, "No options file set, using default values.\n");
  } else
  {
    if (!(fp = fopen (printOpts, "r")))
      fprintf (stderr, "Failed to open file %s, using defaults.\n",
         printOpts);
    else
    {
      /* read options here. Add more for extensions */
      fscanf (fp, "%d", &PRINT_OPTIONS.indexCspaceSupport);
      fclose (fp);
    }
  }

  /* block out SIGPIPE for when clients die */
  act.sa_flags = 0;
  act.sa_handler = SIG_IGN;
  sigaction (SIGPIPE, &act, &oact);

  /* create the accept socket and the file descriptors for the select
     call */
  if ((acceptFd = createAcceptSocket (port)) < 0)
  {
    perror ("create_accept_socket failed");
    return -4;
  }
  highestFd = acceptFd;
  FD_ZERO (&readfds);
  FD_ZERO (&writefds);
  FD_ZERO (&exceptfds);
  FD_ZERO (&readmask);
  FD_SET (acceptFd, &readfds);

  /* initialize the client data structures */
  initPrintClients (PCS, STATICMAXFD);

  eventLoop (readfds, writefds, exceptfds, readmask, acceptFd, highestFd);

	return 0; /* Hush compiler. */
}
	

